//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                         Floppy disk support module                         *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// This modules implements functions to represent the actual physical disks.
// It consists of support functions that manipulate the disk image files.
//
// To do:
//    - Check that seek position is within range of image.  If not then put
//      message out and flag as an error.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.00 - 19 June 2007, S.J.Kay
// The following changes/additions have been made:
// - Added new disk image structures,  this now allows for different floppy
//   images to be used instead of only the 40T DS (SBC) images.
// - disk_open function now uses open with write enabled (r+b). This is
//   required to allow disk_write to work.
// - Added an assert to test max_size variable in disk_read function.
// - Implement the disk_write function.
//
// v0.00 - 7 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "disk.h"
#include "ubee512.h"

//==============================================================================
// Disk open.
//
// Determine the disk format and structure being used.  This can one of the
// following:
//
// * original v0.22 disk format (DS40) Side0=1st 200K, Side1=2nd 200K
// * dsk format
// * dip format (disk image plus) This is an in-house spec not finalised.
//
//   pass: const char *name
//         disk_t *disk
// return: int				0 if no error, else -1 
//==============================================================================
int disk_open (const char *name, disk_t *disk)
{
 dsk_t dsk;			// DSK header structure
 char uname[256];
 
 int i = 0;
 uname[0] = 0; 
 strcat(uname, name);
 while (uname[i])
    {
     uname[i] = toupper(uname[i]);
     i++;
    }

 assert(disk != NULL);

 disk->fdisk = fopen(name, "r+b");

 if (disk->fdisk == NULL)
    return -1;

 if (strstr(uname, ".DIP") != NULL)
    disk->itype = 20;
 else   
    if (strstr(uname, ".DSK") != NULL)
       disk->itype = 10;
    else
       if (strstr(uname, ".IMG") != NULL)
          disk->itype = 0;
       else
          if (strstr(uname, ".NW") != NULL)
             disk->itype = 0;
          else
             return -1;

 if (disk->itype == 0)		// if IMG (nanowasp v0.22 format)
    {
     disk->imagerec.tracks = 40;
     disk->imagerec.heads = 2;
     disk->imagerec.sectrack = 10;
     disk->imagerec.secsize = 512;           
     disk->wrprot = 0;
#ifdef use_info_disk_open
     printf("Image format is RAW 40T DS\n");
#endif
    } 

 if (disk->itype == 10)		// if DSK
    {
     fread(&dsk, sizeof(dsk), 1, disk->fdisk);
     disk->imagerec.tracks = dsk.tracks;
     disk->imagerec.heads = dsk.heads; 
     disk->imagerec.sectrack = dsk.spt;
     disk->imagerec.secsize = 128 << dsk.bps;
#ifdef use_info_disk_open
     printf("Image format is DSK\n");
#endif
    } 

 if (disk->itype == 20)		// if DIP (in house spec)
    {
     fseek(disk->fdisk, -sizeof(disk->imagerec), SEEK_END);
     fread(&disk->imagerec, sizeof(disk->imagerec), 1, disk->fdisk);
     if (strcmp(disk->imagerec.id, "DISK-IMAGE-PLUS") != 0)
        return -1;
     if ((disk->imagerec.secsize != 128) && (disk->imagerec.secsize != 256) &&
         (disk->imagerec.secsize != 512) && (disk->imagerec.secsize != 1024))
        return -1;
     disk->wrprot = disk->imagerec.wrprot;       
#ifdef use_info_disk_open
     printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
             disk->imagerec.ver, disk->imagerec.type);
#endif
    }

#ifdef use_info_disk_open
     printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
#endif

 return 0;    
}

//==============================================================================
// Disk close.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
void disk_close(disk_t *disk)
{
 assert(disk != NULL);

 fclose(disk->fdisk);
 disk->fdisk = NULL;
}

//==============================================================================
// Disk create.
//
//   pass: const char *name
//         disk_t *disk
// return: int				-1
//==============================================================================
int disk_create(const char *name, disk_t *disk)
{
 return -1;
}

//==============================================================================
// Disk read.
//
// If the use_imagerec is defined then the method to access the image is
// defined in the imagerec structure and allows for various floppy disk images
// to be defined.
//
//   pass: disk_t *disk
//         char *buf
//         int max_size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_read(disk_t *disk, char *buf, int max_size, int side, int track, int sect)
{
 int trackofs;
 int sectofs;
 int sectuse;
 int dskofs;

 assert(disk != NULL);
 assert(disk->fdisk != NULL);

#ifdef use_debug_disk_read_abort
 printf("disk_read: forcing an abort here\n");
 assert(1 != 1);
#endif

 if (disk->itype == 0)		// if IMG (nanowasp v0.22 format)
    {
     fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
     fread(buf, 512, 1, disk->fdisk);
    }  

 if (disk->itype == 10)		// if DSK
    {
     sectuse = sect-1;

     trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
     sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
     dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

     fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
     fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
    }
    
 if (disk->itype == 20)		// if DIP (in house spec)    
    {
     if (track >= disk->imagerec.datatrack)
        sectuse = sect - disk->imagerec.datasecofs;
     else
        sectuse = sect - disk->imagerec.systsecofs;

     trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
     sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);

     fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
     fread(buf, disk->imagerec.secsize, 1, disk->fdisk);

#ifdef use_info_disk_read
     printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
            disk->imagerec.ver, disk->imagerec.type);
     printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
     printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
     printf("File offset: %d\n\n", trackofs + sectofs);
#endif
    }
 return 1;
}

//==============================================================================
// Disk write.
//
// If the use_imagerec is defined then the method to access the image is
// defined in the imagerec structure and allows for various floppy disk images
// to be defined.
//
// A flush call is used on each sector write.  This may be moved out at a later
// stage when image (floppy ) swapping is implemented.
//
//   pass: disk_t *disk
//         char *buf
//         int size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_write(disk_t *disk, char *buf, int size, int side, int track, int sect)
{
 int trackofs;
 int sectofs;
 int sectuse; 
 int dskofs;
 
 assert(disk != NULL);
 assert(disk->fdisk != NULL);

 if (disk->itype == 0)		// if IMG (nanowasp v0.22 format)
    {
     fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
     fwrite(buf, 512, 1, disk->fdisk);
     fflush(disk->fdisk);
    }  

 if (disk->itype == 10)		// if DSK
    {
     sectuse = sect-1;

     trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
     sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
     dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;
     
     fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);

     fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
     fflush(disk->fdisk);     
    }
    
 if (disk->itype == 20)		// if DIP (in house spec)    
    {
     if (track >= disk->imagerec.datatrack)
        sectuse = sect - disk->imagerec.datasecofs;
     else
        sectuse = sect - disk->imagerec.systsecofs;       

     trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
     sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
     fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);

     fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
     fflush(disk->fdisk);
    }
 return 1;    
}

//==============================================================================
// disk get sector header ???.
//
//   pass:
// return:
//==============================================================================

/* int disk_getsecthdr(...) ?? */

//==============================================================================
// Disk write protected query.
//
// To add:  Disk images could have an '_' underscore character appended to the
// end of the filename to make the images write protected.  This would mean that
// system dependent permissions would not be require and would be portable.
//
//   pass: disk_t *disk
// return: int				write protect status
//==============================================================================
int disk_iswrprot(disk_t *disk)
{
   if (disk == NULL)
      return 0;

   return disk->wrprot;
}
