//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                              uBee512 module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Provides init, application loop, and deinit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.00 - 8 August 2007, S.J.Kay
// - A major recoding of most of the original source to improve speed
//   regulation and for the addition of new features.
// - Create directories in users home accounts for the placement of disk and
//   ROM images for first time use.
// - Z80_cycles and associated variables now use unsigned 64 bit variables.
// - Added signal call back handler to allow control from other applications.
// - Added video timers and function calls in application loop for blinking
//   cursor and alpha+ hardware video flashing.
// - Added some code in the application loop to support the new sound emulation
//   feature.
// - Recoded the Z80 CPU speed regulation for the clock target.  The code now
//   uses the gettimeofday() function for unices and clock() function in mingw
//   builds to determine delay periods between mz80exec(n) function calls.
// - Added command line option for standard and standard+colour emulation.
// - Added command line monitor type options.
// - Added command line parameters for disk images for drives A-D.
// - Added turbo cammand line option.
// - Added fullscreen command line option.
// - Report init failures, not successes.
//
// v0.00 - 16 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#ifdef MINGW
#else
#include <sys/types.h> 		// various type definitions, like pid_t
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>		// signal name macros, and the signal() prototype
#endif

#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "mz80.h"
#include "gui.h"
#include "pio.h"
#include "sound.h"
#include "paknet.h"

static int init(void);
static int deinit(void);
static int reset(void);

#define LOGFILE	"ubee512_log.txt"

//==============================================================================
// structures and variables
//==============================================================================
FILE *debuglog;
int  debuglvl;

uint64_t z80_cycles;

int key_quit;
int key_reset;

int fullscreen;
int standard;

int flashvideo;

int done;
int loop;

#ifdef MINGW
clock_t ticks1, ticks2;
#else
static struct timeval tod;
uint64_t ticks_tod;
#endif

int delay;

char userhome[512];
char userhome_diskpath[512];
char userhome_romspath[512];

char romimage1[512];
char romimage2[512]; 
char romimage3[512];
char romimage4[512]; 

char floppya[512];
char floppyb[512];
char floppyc[512];
char floppyd[512];

#ifdef use_audio
clock_t ticksound;
#endif

volatile char gui_signal;

static int turbo;

char *floppy_a = NULL;
char *floppy_b = NULL;
char *floppy_c = NULL;
char *floppy_d = NULL;

char *opt_monitor = NULL;
char *opt_colour = NULL;

extern int extendram;
extern int crtcresized;

extern int cur_blink;
extern int cur_mode;
extern int cur_pos;

extern int audio_silence;
extern int dobuffering;

#ifdef MINGW
#else
//==============================================================================
// External GUI signal handler.
//
// This allows an optional external GUI program to change values during the
// running of the emulator.  The code here is kept as simple as possible and
// simply just sets a flag to indicate that a signal was received.  The main
// loop checks the flag value to see if any action is required.
//
// Can be tested in unices by using the following to cause a reset():
// $kill -s 30 pid
//
// To find the pid use:
// $ps ax | grep ubee512
//
//   pass: int sig_num
// return: void
//==============================================================================
void gui_signal_handler (int sig_num)
{
 gui_signal = 1;
}
#endif

//==============================================================================
// Disks and ROMS.
//
//   pass: void
// return: void
//==============================================================================
void disk_and_roms (void)
{
#if MINGW
#define SLASHCHAR '\\'
 strcat(userhome_diskpath, "c:\\ubee512\\disks\\"); 
 strcat(userhome_romspath, "c:\\ubee512\\roms\\");
#else
#define SLASHCHAR '/'
 char *s;
 int dirres;
 mode_t mode;
 
 s = getenv("HOME");
 if (s == NULL)
    {
     printf("uBee512: Unable to get the user home path\n");
     exit(1);
    }
 else
    strcat(userhome, s);

 strcat(userhome, "/.ubee512"); 

 strcat(userhome_diskpath, userhome); 
 strcat(userhome_diskpath, "/disks/");

 strcat(userhome_romspath, userhome);
 strcat(userhome_romspath, "/roms/");

 mode = S_IRWXG | S_IRWXU | S_IROTH | S_IXOTH;
 
 dirres = mkdir(userhome, mode);
 if (dirres == 0)	// if there was no directory and it was created ok
    {
     dirres = mkdir(userhome_diskpath, mode);
     dirres = mkdir(userhome_romspath, mode);
     
     printf("\n");
     printf("=========================================================================\n");
     printf("This is the first time you have run the ubee512 emulator in this account.\n");
     printf("Directories have been created in your home account for first time use.\n");
     printf("\n");
     printf("You now need to copy ROMs charrom.bin and rom1.bin into:\n");
     printf("%s\n", userhome_romspath);
     printf("and disk image(s) to:\n");
     printf("%s\n", userhome_diskpath);     
     printf("\n");     
     printf("When you have done this run the ubee512 emulator again.\n");          
     printf("=========================================================================\n");
     exit(0);		// exit so user can copy the required files
    } 

#endif

// Floppy A image
 if (floppy_a != NULL)
    {
     if (strchr(floppy_a, SLASHCHAR) == NULL)
        {
         floppya[0] = 0;
         strcat(floppya, userhome_diskpath);
         strcat(floppya, floppy_a); 
        }
     else    
        strcat(floppya, floppy_a);
    }
 else
    {
     floppya[0] = 0; 
     strcat(floppya, userhome_diskpath);
     strcat(floppya, "boot.dip");
    } 

// Floppy B image
 if (floppy_b != NULL)
    {
     if (strchr(floppy_b, SLASHCHAR) == NULL)
        {
         floppyb[0] = 0;
         strcat(floppyb, userhome_diskpath);      
         strcat(floppyb, floppy_b); 
        }
     else    
        strcat(floppyb, floppy_b);
    }

// Floppy C image
 if (floppy_c != NULL)
    {
     if (strchr(floppy_c, SLASHCHAR) == NULL)
        {
         floppyc[0] = 0;
         strcat(floppyc, userhome_diskpath);      
         strcat(floppyc, floppy_c); 
        }
     else    
        strcat(floppyc, floppy_c);
    }

// Floppy D image
 if (floppy_d != NULL)
    {
     if (strchr(floppy_d, SLASHCHAR) == NULL)
        {
         floppyb[0] = 0;
         strcat(floppyd, userhome_diskpath);      
         strcat(floppyd, floppy_d); 
        }
     else    
        strcat(floppyd, floppy_d);
    }
    
 printf("Floppy A: %s\n", floppya);
 if (strlen(floppyb))
    printf("Floppy B: %s\n", floppyb);
 if (strlen(floppyc))
    printf("Floppy C: %s\n", floppyc);
 if (strlen(floppyd))
    printf("Floppy D: %s\n", floppyd);

// ROMs 
 strcat(romimage1, userhome_romspath); 
 strcat(romimage1, "rom1.bin"); 

 strcat(romimage2, userhome_romspath); 
 strcat(romimage2, "rom2.bin"); 

 strcat(romimage3, userhome_romspath); 
 strcat(romimage3, "rom3.bin"); 

 strcat(romimage4, userhome_romspath); 
 strcat(romimage4, "charrom.bin"); 
}

//==============================================================================
// Initialise.
//
//   pass: void
// return: int				0 if no error, -1 if error
//==============================================================================
static int init (void)
{
 SDL_Surface *icon;

 disk_and_roms();
 
#ifdef MINGW
#else
// set the GUI signal handler
 signal(30, gui_signal_handler);
#endif

 if (debuglvl > 0)
    {
     debuglog = fopen(LOGFILE, "w"); 
     if (debuglog == NULL)
        {
         fprintf(stderr, "unable to write to log file\n");
         return -1;
        }
     fprintf(debuglog, "debug level = %i\n", debuglvl);
    }

 if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) != 0)
    {
     fprintf(stderr, "init: SDL_Init failed - %s\n", SDL_GetError());
     return -1;
    }

 SDL_EventState(SDL_ACTIVEEVENT | SDL_KEYUP | SDL_KEYDOWN | SDL_SYSWMEVENT, SDL_IGNORE);

 SDL_WM_SetCaption(TITLESTRING, ICONSTRING);

 if ((icon = SDL_LoadBMP("uBee512.bmp")) != NULL)
    {
     SDL_WM_SetIcon(icon, NULL);
     SDL_FreeSurface(icon);
    }

 if (gui_init() != 0)
    {
     printf("Failed: gui_init\n");    
     return -1;
    } 
 else
    {
    }   

 if (z80_init() != 0)
    {
     printf("Failed: z80_init\n");    
     return -1;
    } 

 if (memmap_init() != 0)
    {
     printf("Failed: memmap_init\n");    
     return -1;
    } 

 if (crtc_init() != 0)
    {
     printf("Failed: crtc_init\n");
     return -1;
    }

 if (keyb_init() != 0)
    {
     printf("Failed: keyb_init\n");
     return -1;
    } 

 if (vdu_init() != 0)
    {
     printf("Failed: vdu_init\n");
     return -1;
    } 

 if (fdc_init() != 0)
    {
     printf("Failed: fdc_init\n");
     return -1;
    } 

 if (pio_init() != 0)
    {
     printf("Failed: pio_init\n");
     return -1;
    } 

 if (paknet_init() != 0)
    {
     printf("Failed: paknet_init\n");
     return -1;
    } 

 return 0;
}

//==============================================================================
// De-initialise.
//
//   pass: void
// return: int			result of de-init functions
//==============================================================================
static int deinit(void)
{
 int res;

 if (debuglvl > 0)
    fclose(debuglog);

 res = pio_deinit() || paknet_deinit() || fdc_deinit() ||
       vdu_deinit() || keyb_deinit() || crtc_deinit() || memmap_deinit() ||
       z80_deinit() || gui_deinit();

 SDL_Quit();

 return res;
}

//==============================================================================
// Reset. Reset the virtual Microbee.
//
// Make sure all reset functions clear variables that are constantly active
// values.
//
//   pass: void
// return: int			result of reset functions
//==============================================================================
static int reset (void)
{
 z80_cycles = 0;
 done = 0;
 loop = 0;

#ifdef MINGW
     ticks1 = clock();
#else
     gettimeofday(&tod, NULL);          
     ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif
 
 return z80_reset() || memmap_reset() || crtc_reset() || keyb_reset() ||
        vdu_reset() || fdc_reset() || pio_reset() || paknet_reset();
}

//==============================================================================
// Application setup.
//
//   pass: void
// return: void
//==============================================================================
void ApplicationSetup (void)
{
#ifdef XTAL3375
// For 3.375 MHz XTAL
#define Z80XTAL 3375000L
// number of 112500 cycle loops for exactly 1 second
#define Z80LOOPS 30
// number of cycles for 1/30 of Z80XTAL
#define Z80CYCLES Z80XTAL / Z80LOOPS
// target excution time for 1/30 of Z80XTAL
#define Z80MS (int)(((float)Z80CYCLES/Z80XTAL) * 1000)
#define Z80US (int)(((float)Z80CYCLES/Z80XTAL) * 1000000L)

#else

// For 2 MHz XTAL
#define Z80XTAL 2000000L
// number of X cycle loops for exactly 1 second
#define Z80LOOPS 50
// number of cycles for 1/50 of Z80XTAL
#define Z80CYCLES Z80XTAL / Z80LOOPS
// target excution time for 1/30 of Z80XTAL
#define Z80MS (int)(((float)Z80CYCLES/Z80XTAL) * 1000)
#define Z80US (int)(((float)Z80CYCLES/Z80XTAL) * 1000000L)
#endif

 reset();		// resets the virtual Microbee

#ifdef MINGW
#define AUDIOWAIT 3
#define SPEEDADJ 2	// 2=about 6% faster to stop audio break up
 ticks1 = clock();   
#else
#define AUDIOWAIT 3
#define SPEEDADJ 3	// 3=about 9% faster to stop audio break up
 uint64_t ticks_tod;
 gettimeofday(&tod, NULL); 
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif 
}

//==============================================================================
// Application loop.
//
// Video timers
// ------------
// The CRTC blinking cursor and Alpha+ hardware inverse and flashing video
// timers use the accumulated Z80 cycle count,  this works well as the Z80 CPU
// execution is regulated with delays and so the instruction count is accurate
// over a period of time,  the rate will increase in turbo mode.
//
//   pass: void
// return: void
//==============================================================================
void ApplicationLoop (void)
{
 int dwResult;
 struct mz80context z80; 

 int cur_blink_rate;
 int cur_blink_last=0;
 int flashvideo_last=0;

 while (! done)
    {
#ifdef MINGW
     ticks1 = clock();
#else
     gettimeofday(&tod, NULL);          
     ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif

// Execute a block (Z80CYCLES) of Z80 instructions and return the result.
// The Z80CYCLES value used has been specially selected for timing purposes.
     dwResult = mz80exec(Z80CYCLES);

// If the result is not 0x80000000, it's an address where an invalid
// instruction was hit.
   if (dwResult != 0x80000000)
        {
         mz80GetContext(&z80);
         printf("Invalid instruction at: 0x%4x\n", (unsigned int)z80.z80pc);
         printf("           Clock ticks: %ld\n", z80.z80clockticks);
//
// There appears to be a bug in v3.4 of makez80 or an uninitialised area of
// memory in another module.  It consistently occurs at 0xdec1 after
// 12,210,520 cycles are executed and a reset is then needed to fix the
// situation. The error occurs regardless of what CP/M OS is booted, the
// error still occurs when filling the boot track with 0s and a "jr $". Have
// found this to be a problem in win32 build but may be present on other
// builds too. There may be a variable not initilised correctly and only
// showing up in the win32 build. an old v2.7 of this code appears OK in
// this area but has other problems in itself.  The work around for now is
// to just reset and restart on each invalid instruction encountered.  This
// is NOT the end solution as some un-intened z80 instructions could get
// executed and may not even get an invalid instruction. It was found that
// the minimum work around required the following but decided to do a full
// reset instead. The minimum work around should give some clues.  A bank
// select issue could be causing rubbish to be exectued.
//
// z80_cycles = 0;   
// z80_reset();
// memmap_reset();

         z80_cycles = 0;   
         reset();
//         exit(1);
        }

     z80_cycles += mz80GetElapsedTicks(1);
     
//   printf("z80_cycles=%d\n", z80_cycles);

// Resize the video when certain CRTC registers have been changed under Z80
// emulation. A wait period is introduced here to prevent the display being
// resized too often in a short time period.
     if (crtcresized)
        {
         if (crtcresized++ == 2)	// will be the 2nd time around loop
            crtc_videochange();		// resets crtcresized
        }

// Determine the current status for the CRTC blinking cursor and refresh it
// if this has changed.
     if (cur_mode > 1)			// only if blinking cursor
        {
         if (cur_mode == 2)
            cur_blink_rate = Z80XTAL/2; // blinking at 1/32 field rate
         else   
            cur_blink_rate = Z80XTAL/4; // blinking at 1/16 field rate        

         if ((z80_cycles / cur_blink_rate) & 0x01)
            cur_blink = 0xFF;
         else
            cur_blink = 0x00;

         if (cur_blink != cur_blink_last)
            {
             cur_blink_last = cur_blink;            
             crtc_redraw_char(cur_pos);
            }
        }    

// Determine the current state of the alpha+ flashing video and refresh it
// if this has changed.
     if (extendram)			// only if extended RAM selected
        {
         if ((z80_cycles / (Z80XTAL/4)) & 0x01)
            flashvideo = 1;
         else
            flashvideo = 0;
         
         if (flashvideo != flashvideo_last)
            {
             flashvideo_last = flashvideo;            
             if (extendram)		// only update if extended RAM selected
                crtc_set_redraw();	// force crtc to redraw screen
            }
        }    

#if use_audio
// Sound buffering.  This is required as CPU emulation speed is affected by
// other processes occuring at the same time, we need to try to ensure the
// sound is slightly in front so that it does not matter if the Z80 lags
// behind slightly in creating the sound.  A 99-133mS delay has been found
// do be about right.  If the buffer delay is too long the lag will be
// noticable by the user.  If too short then break up, and crackling will be
// evident.
     if (dobuffering)
        {
         if (ticksound == -1)		// if just starting the buffering
            ticksound = 0;
         else
            {   
             ticksound++;		// sound buffer delay (33mS per inc)
             if (ticksound == AUDIOWAIT) // about 99mS (if 3) + (0-33mS)
                sound_unsilence();	// allow sound to play
            }
        }    
#endif

// Update the display (only if it's update flag is set)           
     gui_redraw();

// Check keys,  if the keys do not get checked in the Z80 code then we must make
// sure it gets checked periodically for the emulator functionality keys.
     keyb_keysupdate();

// Handle external GUI signals
     if (gui_signal)
        {
         reset();
         gui_signal = 0;
        }

// Check for exit and reset key flags
     if (key_quit)
        {
         key_quit = 0;
         done = 1;
        } 
     if (key_reset)
        {
         key_reset = 0;
         reset();
        }

// Insert delays to achieve correct emulated Z80 CPU speed.  The Z80 speed
// may be made a little faster to help prevent audio break up.
//
// The MINGW build relies on a 1mS timer (1ms granularity) and the delay
// required is calculated from the last block (Z80CYCLES) of Z80
// instructions.
//
// The non MINGW builds (unices) uses the gettimeofday function and this is
// converted to a 64 bit variable with a 1mS resolution.
//
     if (turbo)
        SDL_Delay(0);
     else
        {
#ifdef MINGW
// Find the execution delay required for the last (1/Z80LOOPS)
         ticks2 = clock() - ticks1;
         if (ticks2 >= 0)		// if clock() has not wrapped around
            {                   	// else use the last delay value
             delay = (Z80MS-SPEEDADJ) - ticks2;
//           printf("delay=%d\n", delay);                             
            }
#else
         gettimeofday(&tod, NULL);
//       printf("s=%d  us=%d\n", tod.tv_sec, tod.tv_usec);                 
         delay = (Z80MS-SPEEDADJ) - ((uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000)) - ticks_tod);
//       printf("delays=%d\n", delay);                 
#endif
         
// Execute the delay period.  If a delay is not required then a delay of 0
// is used to give up the CPU for other processes.
         if (delay < 0)
            SDL_Delay(0);            
         else
            SDL_Delay(delay);
        }
    }    
}

//==============================================================================
// Main program start point.
//
//   pass: int argc, char *argv[]	Command line arguments
// return: int				0 if no error, 1 if error
//==============================================================================
int main (int argc, char *argv[])
{
 int opt;

 while ((opt = getopt(argc, argv, "a:b:c:d:m:s:ftz")) != -1)
    {
     switch(opt)
        {
         case 'f': fullscreen++;
                   break;
         case 's': standard++;
                   opt_colour = optarg;                   
                   break;
         case 't': turbo++;
                   break;
         case 'z': debuglvl++;
                   break;
         case 'a': floppy_a = optarg;
                   break;
         case 'b': floppy_b = optarg;
                   break;                   
         case 'c': floppy_c = optarg;
                   break;
         case 'd': floppy_d = optarg;
                   break;         
         case 'm': opt_monitor = optarg;
                   break;                            
         default:  exit(1);
                   break;
        }
    }

 if (init() != 0)
    {
     fprintf(stderr, "Fatal error during initialisation.\n");
     SDL_Quit();
     exit(1);
    }

 ApplicationSetup();
 ApplicationLoop();
 
 if (deinit() != 0)
    {
     fprintf(stderr, "Error while de-initialising.\n");
     exit(1);
    }

 return 0;
}
